from freezegun import freeze_time
import datetime
# import asyncio
import pytest
# import pytest-asyncio
from ..common.api_calls import DirectLineAPI
# from unittest.mock import patch


@pytest.fixture
def message_bot():
    return DirectLineAPI()


@pytest.fixture
def message_bot_with_message():
    bot = DirectLineAPI()
    bot.send_message('Hello')
    return bot


# test _generate_token() generates token, conversation id, calls start_conversation
def test_get_token_saves_token_data(message_bot):
    # bot = DirectLineAPI()
    message_bot._get_token()
    assert '_conversationid' in message_bot.__dict__.keys()
    assert '_token_data' in message_bot.__dict__.keys()


def test_send_message(message_bot):

    res = message_bot.send_message("hello")
    assert res.status_code == 200


# TODO: combine incorrect_conversation_id and invalid_token
def test_send_message_incorrect_conversation_id():
    """Sending a message with an valid token but invalid conversation id will
    return None"""
    bot = DirectLineAPI()
    bot._conversationid = "1"
    # A bad request to the message API returns a None response
    assert not bot.send_message("hello")


def test_send_message_invalid_token():
    """Sending a message with an invalid token but valid conversation id will
    return None"""
    bot = DirectLineAPI()
    invalid_token_data = {
        'token': 1,
        'expires_in': 2,
    }
    invalid_header_token = {
        "Authorization": "Bearer 1",
    }
    bot._headers.update(invalid_header_token)
    bot._token_data.update(invalid_token_data)
    # A bad request to the message API returns a None response
    assert not bot.send_message("hello")


def test_get_response(message_bot_with_message):
    bot = DirectLineAPI()
    send_res = bot.send_message("Hi")
    response = bot.get_response()
    # assert 0
    while response['text'] == 'Hi':
        response = bot.get_response()

    assert "Hello, I'm BGS AI Agent - Non Production, a virtual assistant." in response['text']


    # print('*'*30, 'fires')
    # # assert 0
    # asyncio.sleep(20)
    # print('*'*30, 'fires again')

    # res = message_bot_with_message.get_response()
    # # This is currently the stock response from the AI chatbot
    # assert "Hello, I'm BGS AI Agent" in res['text']


def test_refresh_token():
    with freeze_time('2025-01-01 01:00:01'):
        bot = DirectLineAPI()
        assert bot._token_data['expires_at'] == datetime.datetime(2025, 1, 1, 2, 0, 1)
        with freeze_time('2025-01-01 01:59:01'):
            bot.refresh_token()
            assert bot._token_data['expires_at'] == datetime.datetime(2025, 1, 1, 2, 59, 1)


# test _start_conversation() returns 201
# test send_message returns 200 response
# test send_message no conversation id returns 200 response
# test send_message no conversation id returns 200 response


import requests
from django.conf import settings
import datetime

"""
Start the shell with:
1. python manage.py shell 
Commands to run in shell:
1. from bgs_chat.common.websocket import DirectLineAPI
2. bot = DirectLineAPI()
3. bot._get_token()
4. bot.start_conversation()
Then in any order
5. bot.send_message('hello')
6. bot.get_response()
7. bot.get_all_conversation_activity()
8. bot.get_all_messages()
TODO: listener
TODO: listener to be triggered by start_conversation, should be async/await.
May be easier to implement this in JS? ie. send message, make call to get_response every x seconds
This would only work if the watermark is being used effectively and we only get the latest response back, for the relevant question
"""

TOKEN_URL = settings.DIRECT_LINE_URL
DL_URL = 'https://directline.botframework.com/v3/directline'


def save_token_data(self, json_token_response):
    self.headers = {
            "Authorization": f"Bearer {json_token_response['token']}",
            "Content-Type": "application/json"
        }
    return {
        'token': json_token_response['token'],
        'expires_in': json_token_response['expires_in'],
        'recieved_at': datetime.datetime.now(),
        'expires_at': datetime.datetime.now() + datetime.timedelta(seconds=json_token_response['expires_in'])
    }


class DirectLineAPI(object):
    def __init__(self):
        super().__init__()
        self.headers = {
            "Content-Type": "appication/json"
        }
        # TODO: pull user id from session;
        # Logged in token?
        self.user_id = "user"
        self.conversation_id = None
        self.stream_url = None
        self.bot_responses = []
        self.watermark = None

    def _get_token(self):
        """
        Get token and conversation ID.
        This does not start a conversation.
        """
        url = TOKEN_URL
        token_respone = requests.get(url)
        if token_respone.status_code == 200:
            json_token_response = token_respone.json()
            self.conversation_id = json_token_response['conversationId']
            self.token_data = save_token_data(self, json_token_response)

    def start_conversation(self):
        url = '/'.join([DL_URL, 'conversations'])
        res = requests.post(url, headers=self.headers)

        if res.status_code != 201:
            raise Exception(f"Failed to start a conversation {res.status_code}: {res.text}")

        json_res = res.json()
        self.conversation_id = json_res['conversationId']
        self.stream_url = json_res['streamUrl']
        return {
            "conversation_id": self.conversation_id,
            "stream_url": self.stream_url
        }

    def send_message(self, text: str):
        # TODO: sanitise text
        # prevent XSS attacks
        # limit message length?
        if not self.conversation_id:
            raise ValueError("No conversation id")

        url = '/'.join([DL_URL, 'conversations', self.conversation_id, 'activities'])
        message_payload = {
            "type": "message",
            "from": {"id": self.user_id},
            "text": text
        }
        res = requests.post(url, headers=self.headers, json=message_payload)

        if res.status_code not in [200, 201]:
            raise Exception(f"Failed to send message, {res.status_code}: {res.text}")

        return res.json()

    def get_response(self):
        if not self.conversation_id:
            raise ValueError("Conversation ID")

        # TODO: move this behind another function
        url = '/'.join([DL_URL, 'conversations', self.conversation_id, 'activities', f'?watermark {self.watermark}' if self.watermark else ''])
        json_res = requests.get(url, headers=self.headers).json()

        activities = json_res['activities']
        # bot_responses = next(i for i in activities if i['type'] == 'message' and i['from'].get('role') == 'bot')

        # this needs some thinking about, but roughly
        # while len(bot_responses) < len(self.bot_responses)... or maybe len(sent_messages?):
        #     json_res = requests.get(url, headers=self.headers).json()
        #     activities = json_res['activities']
        #     bot_responses = next(i for i in activities if i['type'] == 'message' and i['from'].get('role') == 'bot')
            # save the latest watermark

        self.watermark = json_res['watermark']
        try:
            last_message = next(
                i for i in reversed(activities)
                if i['type'] == 'message'
                and i['from'].get('role') == 'bot'
            )
            # This will return only bot responses.
            # TODO:
            # save the response to self.bot_responses
            # while len(res.bot_responses) == len(self.bot_responses):
                # rerun get_responses every 2? seconds
                # return
            # append to self.bot_responses
            bot_response = {
                'timestamp': last_message['timestamp'],
                'text': last_message['text'],
            }
            self.bot_responses.append(bot_response)
            return bot_response
        # Accounts for the scenario where get_response fires and there are
        # no activities of type=message
        # We return a dict to conform with other response types.
        except StopIteration:
            return {"text": "No response found"}

    def get_all_conversation_activity(self):
        if not self.conversation_id:
            raise ValueError("Conversation ID")

        url = '/'.join([DL_URL, 'conversations', self.conversation_id, 'activities'])
        res = requests.get(url, headers=self.headers)
        json_res = res.json()

        return json_res

    def get_all_messages(self):
        if not self.conversation_id:
            raise ValueError("Conversation ID")
        messages = []
        url = '/'.join([DL_URL, 'conversations', self.conversation_id, 'activities'])
        res = requests.get(url, headers=self.headers)
        json_res = res.json()
        for activity in json_res.get('activities', []):
            if activity['type'] == 'message':
                messages.append({
                    'message': activity['text'],
                    'from': activity['from']
                })

        return messages

    def end_conversation(self):
        url = '/'.join([DL_URL, 'conversations'])
        data = {
            "type": "endOfConversation",
        }
        res = requests.post(url, headers=self.headers, json=data)
        return res.json()


from django.http import HttpResponse
from django.shortcuts import render
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from bgs_chat.common.api_calls import DirectLineAPI
import json


# TODO: this currently uses CSRF exempt - should consider replacing or mitigating against CSSA if we're going to use these
@csrf_exempt
def init_conversation(request):
    if request.method == "POST":
        # needed?
        # body = json.loads(request.body)
        bot = DirectLineAPI()
        response = bot.init_conversation()
        return JsonResponse({"response": response}, status=200)

    return JsonResponse({"error": "Method not allowed"}, status=405)


@csrf_exempt
def send_message(request):
    if request.method == "POST":
        decoded = request.body.decode('utf-8')
        body = json.loads(decoded)
        message = body.get('message')

        if not message:
            return JsonResponse({"error": "Missing message"}, status=400)

        bot = DirectLineAPI()
        result = bot.send_message(message)

        return JsonResponse({"status": "sent", "result": result})

    return JsonResponse({"error": "Method not allowed"}, status=405)


def get_messages(request):
    bot = DirectLineAPI()
    messages = bot.get_response()
    return JsonResponse({"messages": messages})


"""
API calls for interacting with the BGS chat
"""
import asyncio
import requests
from django.conf import settings
import datetime

DIRECT_LINE_URL = settings.DIRECT_LINE_URL
CONVERSATION_URL = 'https://directline.botframework.com/v3/directline/conversations'

# TOKENs VS SECRETS
# Also includes info on refreshing tokens
# https://learn.microsoft.com/en-us/azure/bot-service/rest-api/bot-framework-rest-direct-line-3-0-authentication?view=azure-bot-service-4.0

# The more useful documentation is here, in the response from Henry James
# https://community.powerplatform.com/forums/thread/details/?threadid=b8912aaa-0e8f-4c34-9f39-00ecd95b65ad

"""
Start the shell with:
1. python manage.py shell 
Commands to run in shell:
1. from bgs_chat.common.api_calls import DirectLineAPI
2. bot = DirectLineAPI()
3. bot.send_message('hello')
4. bot.get_response() (this will return an error message until the agent is published)
5. bot.refresh_token()
"""


class DirectLineAPI(object):
    def __init__(self):
        super().__init__()
        self.headers = {
            "Content-Type": "application/json"
        }
        self.init_conversation()
        self._watermark = None
        self.last_message_sent_timestamp = None
        self.last_message_received_timestamp = None

    def save_token_data(self, json_dl_response):
        self.headers.update(
            {"Authorization": f"Bearer {json_dl_response['token']}"}
        )
        return {
            'token': json_dl_response['token'],
            'expires_in': json_dl_response['expires_in'],
            'recieved_at': datetime.datetime.now(),
            'expires_at': datetime.datetime.now() + datetime.timedelta(seconds=json_dl_response['expires_in'])
        }

    def init_conversation(self):
        """
        Get token and conversation ID.
        This does not start a conversation.
        """
        url = DIRECT_LINE_URL
        dl_response = requests.get(url)
        if dl_response.status_code == 200:
            json_dl_response = dl_response.json()
            self._conversationid = json_dl_response['conversationId']
            self._token_data = self.save_token_data(json_dl_response)
            # initiate the conversation
            conversation_url = CONVERSATION_URL
            # NOTE: we should consider how/if we want to log errors here

            conversation_response = requests.post(conversation_url, headers=self.headers)
            if conversation_response.status_code == 201:
                self._start_conversation()
        else:
            return dl_response.status_code

    def _start_conversation(self):
        """
        Start a conversation using the generated conversationId and token
        """
        if not self._conversationid:
            self.init_conversation()
        payload = {
            "type": "event",
            "name": "startConversation"
        }
        url = '/'.join([CONVERSATION_URL, self._conversationid, 'activities'])
        res = requests.post(url, headers=self.headers, json=payload)
        return res.status_code

    def send_message(self, text):
        """Send a message"""
        url = '/'.join([CONVERSATION_URL, self._conversationid, 'activities'])
        # Can optionally include user/from in payload as 'from': {'id': ...},
        message_payload = {
            "conversationId": self._conversationid,
            "type": "message",
            "text": text
        }
        self.last_message_sent_timestamp = datetime.datetime.now()
        res = requests.post(url, headers=self.headers, json=message_payload)
        print('** MESSAGE SENT **')
        print(f'message sent at {self.last_message_sent_timestamp}')
        if res.status_code == 200:
            return res.json()

    async def poll_for_responses(self):
        url = '/'.join([
            CONVERSATION_URL,
            self._conversationid,
            'activities',
            f'?watermark {self._watermark}' if self._watermark else ''
        ])

        res = requests.get(url, headers=self.headers, json={'conversationId': self._conversationid})
        # we need to set self.last_received_message here
        if res.status_code == 200:
            json_res = res.json()
            self._watermark = int(json_res['watermark'])
            activities = json_res['activities']
            try:
                last_message = next(
                    i for i in reversed(activities)
                    if i['type'] == 'message'
                    and i['from'].get('role') == 'bot'
                    )

                last_message_details = {
                    'status_code': res.status_code,
                    'timestamp': last_message['timestamp'],
                    'text': last_message['text'],
                }
                print('** LAST RECEIVED MESSAGE **')
                print(f'{last_message_details=}')
                # self.last_message_received_timestamp = last_message['timestamp']
                return last_message_details
        # Accounts for the scenario where get_response fires and there are
        # no activities of type=message
        # We return a dict to conform with other response types.
            except StopIteration:
                return {"text": "No response found"}

        return {
            'status_code': res.status_code,
            'text': "Error contacting bot for response"
            }

    def get_response(self):
        """Get message response"""
        async def _run():
            while True: 
                response = await self.poll_for_responses()
                if self.last_message_received_timestamp and self.last_message_sent_timestamp and self.last_message_received_timestamp >= self.last_message_sent_timestamp:
                    return response
                await asyncio.sleep(2)

        try:
            return asyncio.run(_run())
        except RuntimeError:
            # safeguards against running inside an existing event loop
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(_run())

        # get_latest_responses:
        # API call to get message history.
        # while last_message_sent_timestamp < last_message_received_timestamp
        # res = self.poll_for_responses()
        # while self.last_message_received_timestamp < self.last_message_sent_timestamp:
            # wait 2
        # res = self.poll_for_responses()
        # import pdb; pdb.set_trace()
        # while self.last_message_received_timestamp < self.last_message_sent_timestamp:
        #     await asyncio.sleep(2)
        #     res = self.poll_for_responses()


        # From the docs:
        # If the client receives an ActivitySet wherein the watermark property
        # is null or missing, it should ignore that value and not overwrite
        # the prior watermark that it received.

    def refresh_token(self):
        # we don't run token_valid here as the token may be expired before refreshing
        url = 'https://directline.botframework.com/v3/directline/tokens/refresh'
        res = requests.post(url, headers=self.headers, json={'conversationId': self._conversationid})
        if res.status_code == 200:
            json_res = res.json()
            if self._conversationid != json_res['conversationId']:
                return "Error refreshing token, conversation IDs do not match"
            new_token_data = save_token_data(self, json_res)
            self._token_data.update(new_token_data)
            return "Token updated"
        return "Error refreshing token"

    def reconnect_to_conversation(self, conversation_id):
        """
        Reconnect to an existing conversation.
        From the docs:
        The watermark parameter is optional. If the watermark parameter is
        specified in the request URI, the conversation replays from the
        watermark, guaranteeing that no messages are lost. If the watermark
        parameter is omitted from the request URI, only messages received
        after the reconnection request are replayed.
        """
        # refresh the token if it has expired.
        # if not token_valid():
        self.refresh_token()

        url = '/'.join([
                CONVERSATION_URL,
                self._conversationid,
                'activities',
                f'?watermark {self._watermark}' if self._watermark else ''
            ])
        res = requests.get(url, headers=self.headers, json={'conversationId': conversation_id})
        if res.status_code == 200:
            json_res = res.json()
            if self._conversationid != json_res['conversationId']:
                return "Error reconnecting to conversation, conversation IDs do not match"
            self.headers.update({'Authorization': f"Bearer {json_res['token']}"})
            return "Reconnected to conversation"
        return "Error reconnecting to conversation"

    def end_conversation(self):
        """End a current conversation"""
        # we still want to make sure the token is valid so we can't
        # erroneously end conversations
        # if token_valid():
        url = '/'.join([CONVERSATION_URL, self._conversationid, 'activities'])
        payload = {
            "type": "endOfConversation",
            "from": {
                "id": "user1"
            }
        }
        requests.post(url, headers=self.headers, json=payload)
            # this returns a conversation id. We could store this, if we chose to.